@@ -306,7 +306,7 @@ impl CertD {
                 }
 
                 let filename = e.file_name();
-                t!("Examining {:?}", filename);
+                t!("Examining {}", String::from_utf8_lossy(filename));
                 if filename.len() != 2 {
                     t!("Wrong length.");
                     continue;
@@ -518,7 +518,7 @@ impl CertD {
 
                     if let Some(revert_to_readir) = revert_to_readir {
                         if revert_to_readir == misses && hits <= 1 {
-                            t!("Too many misses; switching to readir implementation.");
+                            t!("Too many misses; switching to readdir implementation.");
                             return self.tag_readdir_unix();
                         }
                     }
@@ -784,10 +784,41 @@ impl CertD {
         -> Result<(Tag, Option<Vec<u8>>)>
     where
         M: FnOnce(D, Option<&[u8]>) -> Result<MergeResult<'a>>,
+    {
+        self.insert_extended(
+            fingerprint, data, return_inserted,
+            |d| Ok(d),
+            merge,
+            |r| Ok(r))
+    }
+
+    /// Inserts or updates a cert, extended version.
+    ///
+    /// This function is like [`CertD::insert`], but after obtaining
+    /// the lock on the cert-d, and before doing anything else, it
+    /// calls `pre`.  Similarly, just before dropping the lock, it
+    /// calls `post`.
+    ///
+    /// If the `pre` callback returns an error, the operation is
+    /// aborted, and the error is propagated to the caller.  The
+    /// `post` callback can also transform the result.
+    ///
+    /// The caller can use this functionality to get the cert-d's tag
+    /// prior to and after an insert operation, for instance.
+    pub fn insert_extended<'a, D, PRE, M, POST>(
+        &self, fingerprint: &str, data: D,
+        return_inserted: bool, pre: PRE, merge: M, post: POST)
+        -> Result<(Tag, Option<Vec<u8>>)>
+    where
+        PRE: FnOnce(D) -> Result<D>,
+        M: FnOnce(D, Option<&[u8]>) -> Result<MergeResult<'a>>,
+        POST: FnOnce((Tag, Option<Vec<u8>>)) -> Result<(Tag, Option<Vec<u8>>)>,
     {
         let blocking = true;
         self.insert_impl(
-            fingerprint, true, data, return_inserted, merge, blocking)
+            fingerprint, true, data, return_inserted,
+            pre, merge, post,
+            blocking)
     }
 
     /// Inserts or updates a cert, non-blocking variant.
@@ -812,10 +843,40 @@ impl CertD {
         -> Result<(Tag, Option<Vec<u8>>)>
     where
         M: FnOnce(D, Option<&[u8]>) -> Result<MergeResult<'a>>,
+    {
+        self.try_insert_extended(
+            fingerprint, data, return_inserted,
+            |d| Ok(d),
+            merge,
+            |r| Ok(r))
+    }
+
+    /// Inserts or updates a cert, extended version.
+    ///
+    /// This function is like [`CertD::try_insert`], but after
+    /// obtaining the lock on the cert-d, and before doing anything
+    /// else, it calls `pre`.  Similarly, just before dropping the
+    /// lock, it calls `post`.
+    ///
+    /// If the `pre` callback returns an error, the operation is
+    /// aborted, and the error is propagated to the caller.  The
+    /// `post` callback can also transform the result.
+    ///
+    /// The caller can use this functionality to get the cert-d's tag
+    /// prior to and after an insert operation, for instance.
+    pub fn try_insert_extended<'a, D, PRE, M, POST>(
+        &self, fingerprint: &str, data: D,
+        return_inserted: bool, pre: PRE, merge: M, post: POST)
+        -> Result<(Tag, Option<Vec<u8>>)>
+    where
+        PRE: FnOnce(D) -> Result<D>,
+        M: FnOnce(D, Option<&[u8]>) -> Result<MergeResult<'a>>,
+        POST: FnOnce((Tag, Option<Vec<u8>>)) -> Result<(Tag, Option<Vec<u8>>)>,
     {
         let blocking = false;
         self.insert_impl(
-            fingerprint, true, data, return_inserted, merge, blocking)
+            fingerprint, true, data, return_inserted,
+            pre, merge, post, blocking)
     }
 
     /// Inserts or updates a cert.
@@ -839,11 +900,40 @@ impl CertD {
         -> Result<(Tag, Option<Vec<u8>>)>
     where
         M: FnOnce(&'a [u8], Option<&[u8]>) -> Result<MergeResult<'a>>,
+    {
+        self.insert_data_extended(
+            data, return_inserted,
+            |d| Ok(d),
+            merge,
+            |r| Ok(r))
+    }
+
+    /// Inserts or updates a cert, extended version.
+    ///
+    /// This function is like [`CertD::insert_data`], but after
+    /// obtaining the lock on the cert-d, and before doing anything
+    /// else, it calls `pre`.  Similarly, just before dropping the
+    /// lock, it calls `post`.
+    ///
+    /// If the `pre` callback returns an error, the operation is
+    /// aborted, and the error is propagated to the caller.  The
+    /// `post` callback can also transform the result.
+    pub fn insert_data_extended<'a, PRE, M, POST>(
+        &self, data: &'a [u8],
+        return_inserted: bool,
+        pre: PRE, merge: M, post: POST)
+        -> Result<(Tag, Option<Vec<u8>>)>
+    where
+        PRE: FnOnce(&'a [u8]) -> Result<&'a [u8]>,
+        M: FnOnce(&'a [u8], Option<&[u8]>) -> Result<MergeResult<'a>>,
+        POST: FnOnce((Tag, Option<Vec<u8>>)) -> Result<(Tag, Option<Vec<u8>>)>,
     {
         let blocking = true;
         let fingerprint = pgp::fingerprint(data)?;
         self.insert_impl(
-            &fingerprint, true, data, return_inserted, merge, blocking)
+            &fingerprint, true, data, return_inserted,
+            pre, merge, post,
+            blocking)
     }
 
     /// Inserts or updates a cert, non-blocking variant.
@@ -868,12 +958,42 @@ impl CertD {
         -> Result<(Tag, Option<Vec<u8>>)>
     where
         M: FnOnce(&'a [u8], Option<&[u8]>) -> Result<MergeResult<'a>>,
+    {
+        self.try_insert_data_extended(
+            data, return_inserted,
+            |d| Ok(d),
+            merge,
+            |r| Ok(r))
+    }
+
+    /// Inserts or updates a cert, non-blocking variant, extended
+    /// version.
+    ///
+    /// This function is like [`CertD::try_insert_data`], but after
+    /// obtaining the lock on the cert-d, and before doing anything
+    /// else, it calls `pre`.  Similarly, just before dropping the
+    /// lock, it calls `post`.
+    ///
+    /// If the `pre` callback returns an error, the operation is
+    /// aborted, and the error is propagated to the caller.  The
+    /// `post` callback can also transform the result.
+    pub fn try_insert_data_extended<'a, PRE, M, POST>(
+        &self, data: &'a [u8],
+        return_inserted: bool,
+        pre: PRE, merge: M, post: POST)
+        -> Result<(Tag, Option<Vec<u8>>)>
+    where
+        PRE: FnOnce(&'a [u8]) -> Result<&'a [u8]>,
+        M: FnOnce(&'a [u8], Option<&[u8]>) -> Result<MergeResult<'a>>,
+        POST: FnOnce((Tag, Option<Vec<u8>>)) -> Result<(Tag, Option<Vec<u8>>)>,
     {
         let blocking = false;
         pgp::plausible_tsk_or_tpk(data)?;
         let fingerprint = pgp::fingerprint(data)?;
         self.insert_impl(
-            &fingerprint, true, data, return_inserted, merge, blocking)
+            &fingerprint, true, data, return_inserted,
+            pre, merge, post,
+            blocking)
     }
 
     /// Inserts or updates the cert or key stored under a special name.
@@ -908,10 +1028,47 @@ impl CertD {
     ) -> Result<(Tag, Option<Vec<u8>>)>
     where
         M: FnOnce(D, Option<&[u8]>) -> Result<MergeResult<'a>>,
+    {
+        self.insert_special_extended(
+            special_name, data, return_inserted,
+            |d| Ok(d),
+            merge,
+            |r| Ok(r))
+    }
+
+    /// Inserts or updates the cert or key stored under a special
+    /// name, extended version.
+    ///
+    /// This function is like [`CertD::insert_special`], but after
+    /// obtaining the lock on the cert-d, and before doing anything
+    /// else, it calls `pre`.  Similarly, just before dropping the
+    /// lock, it calls `post`.
+    ///
+    /// If the `pre` callback returns an error, the operation is
+    /// aborted, and the error is propagated to the caller.  The
+    /// `post` callback can also transform the result.
+    ///
+    /// The caller can use this functionality to get the cert-d's tag
+    /// prior to and after an insert operation, for instance.
+    pub fn insert_special_extended<'a, D, PRE, M, POST>(
+        &self,
+        special_name: &str,
+        data: D,
+        return_inserted: bool,
+        pre: PRE, merge: M, post: POST
+    ) -> Result<(Tag, Option<Vec<u8>>)>
+    where
+        PRE: FnOnce(D) -> Result<D>,
+        M: FnOnce(D, Option<&[u8]>) -> Result<MergeResult<'a>>,
+        POST: FnOnce((Tag, Option<Vec<u8>>)) -> Result<(Tag, Option<Vec<u8>>)>,
     {
         let blocking = true;
         self.insert_impl(
-            special_name, false, data, return_inserted, merge, blocking)
+            special_name, false, data, return_inserted,
+            pre,
+            merge,
+            post,
+            blocking)
     }
 
     /// Inserts or updates the cert or key stored under a special
@@ -949,21 +1106,62 @@ impl CertD {
     where
         M: FnOnce(D, Option<&[u8]>) -> Result<MergeResult<'a>>,
     {
-        let blocking = false;
-        self.insert_impl(
-            special_name, false, data, return_inserted, merge, blocking)
+        self.try_insert_special_extended(
+            special_name, data, return_inserted,
+            |d| Ok(d),
+            merge,
+            |r| Ok(r))
     }
 
-    fn insert_impl<'a, D, M>(
+    /// Inserts or updates the cert or key stored under a special
+    /// name, non-blocking variant, extended version.
+    ///
+    /// This function is like [`CertD::try_insert_special`], but after
+    /// obtaining the lock on the cert-d, and before doing anything
+    /// else, it calls `pre`.  Similarly, just before dropping the
+    /// lock, it calls `post`.
+    ///
+    /// If the `pre` callback returns an error, the operation is
+    /// aborted, and the error is propagated to the caller.  The
+    /// `post` callback can also transform the result.
+    ///
+    /// The caller can use this functionality to get the cert-d's tag
+    /// prior to and after an insert operation, for instance.
+    pub fn try_insert_special_extended<'a, D, PRE, M, POST>(
+        &self,
+        special_name: &str,
+        data: D,
+        return_inserted: bool,
+        pre: PRE, merge: M, post: POST
+    ) -> Result<(Tag, Option<Vec<u8>>)>
+    where
+        PRE: FnOnce(D) -> Result<D>,
+        M: FnOnce(D, Option<&[u8]>) -> Result<MergeResult<'a>>,
+        POST: FnOnce((Tag, Option<Vec<u8>>)) -> Result<(Tag, Option<Vec<u8>>)>,
+    {
+        let blocking = false;
+        self.insert_impl(
+            special_name, false, data, return_inserted,
+            pre,
+            merge,
+            post,
+            blocking)
+    }
+
+    fn insert_impl<'a, D, PRE, M, POST>(
         &self,
         name: &str, name_is_fingerprint: bool,
         data: D,
         return_inserted: bool,
+        pre: PRE,
         merge: M,
+        post: POST,
         blocking: bool,
     ) -> Result<(Tag, Option<Vec<u8>>)>
     where
+        PRE: FnOnce(D) -> Result<D>,
         M: FnOnce(D, Option<&[u8]>) -> Result<MergeResult<'a>>,
+        POST: FnOnce((Tag, Option<Vec<u8>>)) -> Result<(Tag, Option<Vec<u8>>)>,
     {
         let name = if name_is_fingerprint {
             pgp::canonicalize_fingerprint(name)?
@@ -972,8 +1170,6 @@ impl CertD {
         };
 
         let target_path = self.get_path(&name)?;
-        // Make sure the directory exists.
-        fs::create_dir_all(target_path.parent().expect("at least one leg"))?;
 
         let mut lf = RwLock::new(self.idempotent_create_lockfile()?);
         // Lock exclusively
@@ -983,6 +1179,12 @@ impl CertD {
             lf.try_write()?
         };
 
+        // Let the caller examine the cert-d with the lock.
+        let data = pre(data)?;
+
+        // Make sure the directory exists.
+        fs::create_dir_all(target_path.parent().expect("at least one leg"))?;
+
         let old_cert = self.get(&name)?.map(|(_, cert)| cert);
         let old_cert = old_cert.as_deref();
         let merge_result = merge(data, old_cert)?;
@@ -1012,14 +1214,18 @@ impl CertD {
 
         let tag = fs::File::open(&target_path)?.metadata()?.try_into()?;
 
+        let cert = if return_inserted {
+            Some(new_cert.to_vec())
+        } else {
+            None
+        };
+
+        // Let the caller examine the result with the cert-d lock.
+        let (tag, cert) = post((tag, cert))?;
+
         drop(lock);
 
-        Ok((tag,
-            if return_inserted {
-                Some(new_cert.to_vec())
-            } else {
-                None
-            }))
+        Ok((tag, cert))
     }
 
     /// Iterates over the certs in the store returning their fingerprints.
@@ -1116,6 +1322,7 @@ impl CertD {
         std::fs::OpenOptions::new()
             .write(true)
             .create(true)
+            .truncate(true)
             .open(lock_path)
             .map_err(Into::into)
     }
